"""
sanskrit.util.functions
~~~~~~~~~~~~~~~~~~~~~~~

Various helper functions.

:license: MIT and BSD
"""

from __future__ import print_function
import csv


def read_csv(filename):
    """Read from a CSV file using a DictReader.

    :param filename: the name of the file
    """
    with open(filename, 'r') as f:
        reader = csv.DictReader(f)
        for row in reader:
            yield row


def heading(s, char='-'):
    """Print `s` as a heading. This is used to update the user during a
    long-running function."""
    print()
    print(s)
    print(char * len(s))


def tick(s):
    """Print `s` as a list item. This is used to update the user during a
    long-running function.
    """
    print(' -', s)


def tick_every(n):
    def func(x):
        func.i += 1
        if func.i % n == 0:
            print(func.i, ':', x)
    func.i = 0
    return func



# -*- coding: utf-8 -*-
"""
    sanskrit_util.generate
    ~~~~~~~~~~~~~~~~~

    Generators for various Sanskrit forms.

    Due to its inflectional nature, Sanskrit has tens of millions of
    different forms. It is not always feasible to store them in the
    database. But since the majority of the language is regular, we can
    store the principal parts of some group of forms and generate the
    whole group by applying some basic rules. That is, if we have 2000
    adjective stems and the 72 endings that they use, we can store 2072
    forms instead of 144k forms.

    :license: MIT and BSD
"""
from __future__ import print_function
from . import util
from .schema import NominalEnding


class Generator(object):

    """Template for a generator."""

    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    def paradigm(self, *args, **kwargs):
        """Generate a full paradigm."""
        raise NotImplementedError


class NominalGenerator(Generator):

    """
    A generator for nominal forms.

    :param ctx: some :class:`~sanskrit_util.Context`.
    """

    def __init__(self, ctx):
        self.ctx = ctx
        session = ctx.session

        self.nominal_stem_trie = util.HashTrie()
        self.nominal_endings = {}
        seen = set()
        for e in session.query(NominalEnding):
            stem_type = e.stem_type
            if stem_type not in seen:
                seen.add(stem_type)
                self.nominal_stem_trie[stem_type[::-1]] = stem_type
                self.nominal_endings[stem_type] = {}

            key = (e.gender_id, e.case_id, e.number_id)
            self.nominal_endings[stem_type][key] = e.name
        session.remove()

    def paradigm(self, stem_name, gender):
        """Generate a full paradigm using normal Sanskrit rules. The
        function treats irregular stems as regular.
        """

        stem_types = self.nominal_stem_trie[stem_name[::-1]]
        stem_type = max(stem_types, key=len)
        truncated = stem_name[:-len(stem_type)]
        endings = self.nominal_endings[stem_type]

        returned = {}
        enum = self.ctx.enum_id
        gender_id = enum['gender'][gender]
        for case in '12345678':
            case_id = enum['case'][case]
            for number in 'sdp':
                number_id = enum['number'][number]
                key = (gender_id, case_id, number_id)
                returned[(case, number)] = truncated + endings[key]

        return returned



# -*- coding: utf-8 -*-
"""
    sanskrit_util.query
    ~~~~~~~~~~~~~~~~~~~~~~~

    A simple interface for querying Sanskrit linguistic data.

    :license: MIT and BSD
"""

from collections import defaultdict
import six
from . import sounds
from .generate import NominalGenerator
from .schema import *


class SimpleQuery(object):

    """A simple API for database access."""

    def __init__(self, ctx):
        self.ctx = ctx
        self.session = ctx.session
        self.nominal = NominalGenerator(ctx)

        # Store IDs of irregular stems
        irreg = self.session.query(StemIrregularity) \
                    .filter(StemIrregularity.fully_described == True)

        self.irregular_stems = set([x.id for x in irreg])
        self.session.remove()

    def _fetch_nominal_paradigm(self, stem_id, gender_id):
        """Fetch a nominal paradigm from the database."""
        enum_abbr = self.ctx.enum_abbr

        results = self.session.query(Nominal)\
                              .filter(Nominal.stem_id == stem_id)\
                              .filter(Nominal.gender_id == gender_id)

        returned = {}
        for nominal in results:
            case = enum_abbr['case'][nominal.case_id]
            number = enum_abbr['number'][nominal.number_id]
            returned[(case, number)] = nominal.name

        return returned

    def _nominal_stem(self, stem_name, stem_cls=None):
        """Fetch a nominal stem from the database."""
        stem_cls = stem_cls or Stem
        stem = self.session.query(stem_cls)\
                           .filter(stem_cls.name == stem_name)\
                           .first()
        return stem

    def _simplify(self, forms):
        """Simplify the given forms by applying consonant reduction."""
        for parse, name in six.iteritems(forms):
            forms[parse] = sounds.Term(name).simplify()

    def noun(self, stem_name, gender):
        """Query for nouns.

        :param stem_name: the stem name
        :param gender: the noun gender
        """
        stem = self._nominal_stem(stem_name, NominalStem)
        if stem is None:
            return {}

        if stem.id in self.irregular_stems:
            gender_id = self.ctx.enum_id['gender'][gender]
            returned = self._fetch_nominal_paradigm(stem.id, gender_id)
        elif stem:
            returned = self.nominal.paradigm(stem_name, gender)

        self._simplify(returned)
        return returned

    def pronoun(self, stem_name, gender):
        """Query for pronouns.

        :param stem_name: the stem name
        :param gender: the pronoun gender
        """
        stem = self._nominal_stem(stem_name, PronounStem)
        if stem is None:
            return {}

        gender_id = self.ctx.enum_id['gender'][gender]
        returned = self._fetch_nominal_paradigm(stem.id, gender_id)

        self._simplify(returned)
        return returned

    def verb(self, root_name, mode, voice, vclass=None, **kw):
        """Query for inflected verbs.

        :param root_name: the verb root
        :param mode: the verb mode
        :param voice: the verb voice
        :param vclass: the verb class to use. This can be used to
                       distinguish between homophonous roots, such as
                       'kR' ("do") and 'kR' ("praise").
        """
        enum_id = self.ctx.enum_id
        enum_abbr = self.ctx.enum_abbr
        session = self.session

        roots = session.query(Root).filter(Root.name == root_name)
        roots_id = [r.id for r in roots]
        mode_id = enum_id['mode'][mode]
        voice_id = enum_id['voice'][voice]

        returned = {}
        results = session.query(Verb)\
                         .filter(Verb.root_id.in_(roots_id))\
                         .filter(Verb.mode_id == mode_id)\
                         .filter(Verb.voice_id == voice_id)
        for verb in results:
            person = enum_abbr['person'][verb.person_id]
            number = enum_abbr['number'][verb.number_id]
            returned[(person, number)] = verb.name

        session.close()
        self._simplify(returned)
        return returned

    def verb_summary(self, root_name, vclass=None):
        """Query for a summary of a verb's behavior.

        The function returns the following information:

        - the root ID
        - the 3rd. sg. forms of all verbs produced from the plain root
        - the stems of all participles produced from the plain root

        :param root: the verb root
        :param vclass: the verb class to use. This can be used to
                       distinguish between homophonous roots, such as
                       'kR' ("do") and 'kR' ("praise").
        """
        verbs = defaultdict(list)
        participles = defaultdict(list)

        ctx = self.ctx
        session = self.session

        # abbr -> ID
        ei_person = ctx.enum_id['person']
        ei_number = ctx.enum_id['number']

        # ID -> abbr
        ea_mode = ctx.enum_abbr['mode']
        ea_voice = ctx.enum_abbr['voice']

        # Root
        root = session.query(Root).filter(Root.name == root_name).first()
        if root is None:
            return {}
        root_id = root.id

        # Verbs
        results = session.query(Verb).filter(Verb.root_id == root_id) \
                         .filter(Verb.person_id == ei_person['3']) \
                         .filter(Verb.number_id == ei_number['s'])\

        for r in results:
            mode = ea_mode[r.mode_id]
            voice = ea_voice[r.voice_id]
            verbs[(mode, voice)].append(r.name)

        # Participles
        results = session.query(ParticipleStem) \
                         .filter(ParticipleStem.root_id == root_id)

        for r in results:
            mode = ea_mode[r.mode_id]
            voice = ea_voice[r.voice_id]
            participles[(mode, voice)].append(r.name)

        session.close()
        return {
            'root_id': root_id,
            'verbs': verbs,
            'participles': participles,
        }



# -*- coding: utf-8 -*-
"""
    sanskrit.analyze
    ~~~~~~~~~~~~~~~~

    Code for *analyzing* Sanskrit forms, i.e. finding the basic lexical
    forms that produced them and specifying the word's inflectional
    information.

    :license: MIT
"""

from collections import defaultdict, namedtuple

from . import sounds, util
from .schema import *


Ending = namedtuple('Ending', ['name', 'length', 'stem_type', 'gender_id',
                               'case_id', 'number_id', 'compounded',
                               'is_consonant_stem'])


class Analyzer(object):

    """analyzer"""

    def __init__(self):
        raise NotImplementedError

    def analyze(self, token):
        raise NotImplementedError


class SimpleAnalyzer(Analyzer):

    """A simple analyzer for Sanskrit words. The analyzer is simple
    for a few reasons:

    - It doesn't do any caching.
    - It uses an ORM instead of raw SQL queries.
    - Its output is always "well-formed." For example, neuter nouns can
      take only neuter endings.

    This analyzer is best used when memory is at a premium and speed is
    a secondary concern (e.g. when on a web server).
    """

    def __init__(self, ctx):
        self.ctx = ctx
        self.session = ctx.session

        self.nominal_endings = util.HashTrie()
        for e in self.session.query(NominalEnding):
            stem_type = e.stem_type
            is_cons = stem_type[-1] in sounds.CONSONANTS
            if e.stem_type == "_": 
                e.stem_type = ""
                is_cons = True

            data = {
                'name': e.name,
                'stem_type': e.stem_type,
                'length': len(e.name),
                'gender_id': e.gender_id,
                'case_id': e.case_id,
                'number_id': e.number_id,
                'compounded': e.compounded,
                'is_consonant_stem': is_cons,
            }
            self.nominal_endings[e.name[::-1]] = Ending(**data)
            if 'n' in e.name:
                # TODO: do this more rigorously
                reversed_name = e.name.replace('n', 'R')
                data['name'] = reversed_name
                self.nominal_endings[reversed_name[::-1]] = Ending(**data)


        self.session.remove()

    def _analyze_as_form(self, word):
        """
        Analyze a word by searching for an exact match in the database.

        :param word: the word to analyze
        """
        session = self.session
        results = session.query(Form).filter(Form.name == word).all()
        return results

    def _analyze_as_stem(self, word):
        """
        Analyze a word by searching for the nominal stems that might
        have produced it.

        :param word: the word to analyze
        """
        session = self.session
        gender_set = self.ctx.gender_set
        returned = []

        # Find all stems that could produce this word. Some of these
        # stems might not exist.
        stem_endings_map = defaultdict(set)
        endings = self.nominal_endings[word[::-1]]
        for e in endings:
            truncated_stem = word[:-e.length] or word
            if e.is_consonant_stem:
                # Stem must exist and end in a consonant.
                if not truncated_stem:
                    continue
                if truncated_stem[-1] in sounds.VOWELS:
                    continue
                if truncated_stem in sounds.CONSONANTS:
                    continue
                stem = truncated_stem
            else:
                stem = truncated_stem + e.stem_type

            stem_endings_map[stem].add(e)

        if not stem_endings_map:
            return []

        # Check which of these stems are viable
        stems = session.query(Stem) \
                       .filter(Stem.name.in_(stem_endings_map.keys()))

        # Reattach endings to viable stems
        for stem in stems:
            name = stem.name

            # For nouns, disregard endings that don't match the stem's
            # genders.
            # TODO: fix semantics of this
            if stem.pos_id == Tag.NOMINAL:
                stem_genders = gender_set[stem.genders_id]
                endings = (e for e in stem_endings_map[name]
                           if e.gender_id in stem_genders)
            else:
                endings = stem_endings_map[name]

            for e in endings:
                datum = {
                    'name': word,
                    'pos_id': stem.pos_id,
                    'stem': stem,
                    'gender_id': e.gender_id,
                    'case_id': e.case_id,
                    'number_id': e.number_id,
                    'compounded': e.compounded,
                }
                returned.append(Nominal(**datum))

        return returned

    def analyze(self, word):
        """Return all possible solutions for the given word. Any ORM
        objects used in these solutions will be in a detached state.

        :param word: the word to analyze. This should be a complete
                     word, or what Panini would call a *pada*.
        """
        returned = self._analyze_as_form(word)
        returned.extend(self._analyze_as_stem(word))
        return returned



import heapq


class PriorityQueue(object):

    """A priority queue. Higher values are popped first."""

    def __init__(self):
        self.heap = []

    def __copy__(self):
        p = PriorityQueue()
        p.heap = list(self.heap)

    def __str__(self):
        return str([(item, priority) for priority, item in self.heap])

    def __len__(self):
        return len(self.heap)

    def __iter__(self):
        for priority, item in self.heap:
            yield (item, -priority)

    def push(self, item, priority=0):
        """Add an item to the queue.

        :param item: the item to add
        :param priority: the priority to use
        """
        pair = (-priority, item)
        heapq.heappush(self.heap, pair)

    def pop(self):
        """Pop the highest-priority item from the queue."""
        return heapq.heappop(self.heap)[-1]

    def pop_with_priority(self):
        """Pop the highest-priority item and its priority from the queue::

            item, priority = q.pop_with_priority()

        ::
        """
        (priority, item) = heapq.heappop(self.heap)
        return (item, -priority)

    def peek(self):
        """Read the highest-priority item and its priority without removing
        them from the queue::

            item, priority = q.peek()
            assert (item, priority) == q.peek()

        If the queue is empty, throw an :exc:`IndexError`.
        """
        priority, item = self.heap[0]
        return (item, -priority)



"""
    sanskrit_util.models
    ~~~~~~~~~~~~~~~

    Statistical models.

    :license: MIT
"""
from __future__ import division
from builtins import next, zip, range
import collections
import itertools
import math


#: Marks the beginning or end of a sequence.
SEQUENCE_BOUNDARY = '__SEQ_BOUNDARY__'


class SequenceModel(object):

    """A bigram model over sequences of events."""

    def __init__(self):
        self.prior = collections.Counter()
        self.joint = collections.Counter()
        self.num_prior_events = 80   #TODO: don't hard code

    def _iter_ngrams(self, seq, size):
        iters = itertools.tee(seq, size)
        for i in range(1, size):
            for each in iters[i:]:
                next(each, None)
        return zip(*iters)

    def insert(self, seq):
        """
        :param seq: a list of events
        """
        seq = itertools.chain([SEQUENCE_BOUNDARY], seq, [SEQUENCE_BOUNDARY])
        for x, y in self._iter_ngrams(seq, 2):
            self.joint[(x, y)] += 1
            self.prior[x] += 1

    def log_cond_prob(self, xs, y, delta=1):
        """Return log(P(y | xs)) with add-`delta` smoothing.

        P(y | xs) is approximated as P(y | xs[-1]).

        :param y: the posterior event
        :param xs: all prior events
        """
        x_n = xs[-1] if len(xs) else None
        numerator = self.joint[(x_n, y)] + delta
        denominator = self.prior[x_n] + self.num_prior_events
        assert denominator > 0
        return math.log(numerator / denominator)


class FeatureModel(object):

    """Assigns a score using a small set of features."""

    def score(self, item, remainder):
        is_form = item.form.__class__.__name__ != 'NonForm'
        if is_form:
            item_len = len(item.form.name) if is_form else 0
            finished_chunk = 0 if remainder else 1
            return item_len + finished_chunk
        else:
            return 0



# -*- coding: utf-8 -*-
"""
    sanskrit_util.schema
    ~~~~~~~~~~~~~~~

    Schema for Sanskrit data.
"""

import re

from sqlalchemy import Boolean, Column, ForeignKey, Integer, String
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship


Base = declarative_base()


class SimpleBase(Base):

    """A simple default base class.

    This automatically creates:

    - __tablename__
    - id (primary key)
    - name (string)
    """

    __abstract__ = True

    @declared_attr
    def __tablename__(cls):
        return re.sub('(?<!^)(?=[A-Z])', '_', cls.__name__).lower()

    id = Column(Integer, primary_key=True)
    name = Column(String, index=True)

    def __repr__(self):
        cls = self.__class__.__name__
        return "%s%r" % (cls, (self.id, self.name))


class EnumBase(SimpleBase):

    """Base class for enumerations.

    Each enumeration has a name and an abbreviation.
    """

    __abstract__ = True
    abbr = Column(String)

    def __repr__(self):
        cls = self.__class__.__name__
        return "%s%r" % (cls, (self.id, self.name, self.abbr))


# Enumerations
# ------------
# Instead of enums, we use the following:

class Person(EnumBase):

    """Grammatical person:

    - first person, corresponding to Panini's **uttamapuruṣa**
    - second person, corresponding to Panini's **madhyamapuruṣa**
    - third person, corresponding to Panini's **prathamapuruṣa**
    """


class Number(EnumBase):

    """Grammatical number:

    - singular, corresponding to Panini's **ekavacana**
    - dual, corresponding to Panini's **dvivacana**
    - plural, corresponding to Panini's **bahuvacana**
    """


class Mode(EnumBase):

    """Tenses and moods:

    - present, corresponding to Panini's **laṭ**
    - aorist, corresponding to Panini's **luṅ**
    - imperfect, corresponding to Panini's **laṅ**
    - perfect, corresponding to Panini's **liṭ**
    - simple future, corresponding to Panini's **lṛṭ**
    - distant future, corresponding to Panini's **luṭ**
    - conditional, corresponding to Panini's **lṛṅ**
    - optative, corresponding to Panini's **vidhi-liṅ**
    - imperative, corresponding to Panini's **loṭ**
    - benedictive, corresponding to Panini's **āśīr-liṅ**
    - injunctive
    - future optative
    - future imperative
    """


class Voice(EnumBase):

    """Grammatical voice:

    - parasmaipada
    - ātmanepada
    - ubhayapada
    """


class Gender(EnumBase):

    """Grammatical gender:

    - masculine, corresponding to Panini's **puṃliṅga**
    - feminine, corresponding to Panini's **strīliṅga**
    - neuter, corresponding to Panini's **napuṃsakaliṅga**
    - unknown/undefined
    """


class GenderGroup(EnumBase):

    """Grammatical gender of a nominal stem. Since stems support nearly
    every combination of genders, this class stores both individual
    genders:

    - masculine
    - feminine
    - neuter
    - unknown/undefined

    and collections of genders:

    - masculine and feminine
    - masculine and neuter
    - feminine and neuter
    - masculine, feminine, and neuter
    """

    assocs = relationship('GenderGroupAssociation')
    members = association_proxy('assocs', 'gender')


class Case(EnumBase):

    """Grammatical case.

    - Nominative case, corresponding to Panini's **prathamā**
    - Accusative case, corresponding to Panini's **dvitīyā**
    - Instrumental case, corresponding to Panini's **tṛtīyā**
    - Dative case, corresponding to Panini's **caturthī**
    - Ablative case, corresponding to Panini's **pañcamī**
    - Genitive case, corresponding to Panini's **ṣaṣṭhī**
    - Locative case, corresponding to Panini's **saptamī**
    - Vocative case, corresponding to Panini's **saṃbodhana**
    """


class VClass(EnumBase):

    """Verb class:

    - class 1, corresponding to Panini's **bhvādi**
    - class 2, corresponding to Panini's **adādi**
    - class 3, corresponding to Panini's **juhotyādi**
    - class 4, corresponding to Panini's **divādi**
    - class 5, corresponding to Panini's **svādi**
    - class 6, corresponding to Panini's **tudādi**
    - class 7, corresponding to Panini's **rudhādi**
    - class 8, corresponding to Panini's **tanādi**
    - class 9, corresponding to Panini's **kryādi**
    - class 10, corresponding to Panini's **curādi**
    - class unknown, for verbs like "ah"
    - nominal, corresponding to various Paninian terms
    """

    __tablename__ = 'vclass'


class Modification(EnumBase):

    """Verb modification:

    - Causative, corresponding to Panini's **ṇic**
    - Desiderative, corresponding to Panini's **san**
    - Intensive, corresponding to Panini's **yaṅ**
    """


# Roots, stems, and forms
# =======================

class Tag(SimpleBase):

    """Part of speech tag. It contains the following:

    - noun
    - pronoun
    - adjective
    - indeclinable
    - verb
    - gerund
    - infinitive
    - participle
    - perfect indeclinable (also known as the *periphrastic perfect*)
    - noun prefix
    - verb prefix
    """

    VERB = 1
    NOMINAL = 2
    PRONOUN = 3
    PARTICIPLE = 4
    INDECLINABLE = 5
    VERBAL_INDECLINABLE = 6
    GERUND = 7
    INFINITIVE = 8
    PERFECT_INDECLINABLE = 9
    NOUN_PREFIX = 10
    VERB_PREFIX = 11


# Unfinished forms
# ----------------
# Roots, stems, and prefixes

class Prefix(SimpleBase):

    """A generic prefix."""

    pos_id = Column(ForeignKey(Tag.id))

    pos = relationship(Tag)
    __mapper_args__ = {'polymorphic_on': pos_id}


class VerbPrefix(Prefix):

    """A verb prefix. This corresponds to Panini's **gati**, which includes
    **cvi** (`svāgatī-karoti`) and **upasarga** (`anu-karoti`).
    """

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.VERB_PREFIX}


class NounPrefix(Prefix):

    """A noun prefix. This includes `nañ`, among others."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.NOUN_PREFIX}


class NominalEnding(SimpleBase):

    """A suffix for regular nouns and adjectives. This corresponds to
    Panini's **sup**."""

    CONSONANT_STEM_TYPE = '$cons'

    stem_type = Column(String)
    gender_id = Column(ForeignKey(Gender.id))
    case_id = Column(ForeignKey(Case.id))
    number_id = Column(ForeignKey(Number.id))
    compounded = Column(Boolean)

    gender = relationship(Gender)
    case = relationship(Case)
    number = relationship(Number)


class VerbEnding(SimpleBase):

    """A suffix for conjugated verbs of any kind. This corresponds to
    Panini's **tiṅ**."""

    #: Name of the stem category that uses this ending. This is useful
    #: for partitioning the list of endings according to the stem under
    #: consideration.
    #:
    #: The names of these groups depend on the initial data. The default
    #: data uses these names:
    #: - `'simple'` for classes 1, 4, 6, and 10
    #: - `'complex'` for classes 2, 3, 45, 7, 8, and 9
    #: - `'both'` for all classes
    category = Column(String)
    person_id = Column(ForeignKey(Person.id))
    number_id = Column(ForeignKey(Number.id))
    mode_id = Column(ForeignKey(Mode.id))
    voice_id = Column(ForeignKey(Voice.id))

    person = relationship(Person)
    number = relationship(Number)
    mode = relationship(Number)
    voice = relationship(Voice)


class Root(SimpleBase):

    """A verb root. This corresponds to Panini's **dhātu**:

        | 1.3.1 "bhū" etc. are called `dhātu`.
        | 3.1.22 Terms ending in `san` etc. are called dhātu.

    Moreover, :class:`Root` contains prefixed roots. Although this modeling
    choice is non-Paninian, it does express the notion that verb prefixes can
    cause profound changes in a root's meaning and identity.
    """

    # re-declared for use with `remote_side` below
    id = Column(Integer, primary_key=True)

    #: The ultimate ancestor of this root. For instance, the basis of
    #: "sam-upa-gam" is "gam".
    basis_id = Column(ForeignKey('root.id'))
    discriminator = Column(Integer)

    basis = relationship('Root', remote_side=[id])
    paradigms = relationship('Paradigm')
    vclasses = association_proxy('paradigms', 'vclass')
    voices = association_proxy('paradigms', 'voice')
    __mapper_args__ = {'polymorphic_on': discriminator}


class PrefixedRoot(Root):

    """A root with one or more prefixes."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': 1}

    prefix_assocs = relationship('RootPrefixAssociation',
                                 collection_class=ordering_list('position'),
                                 order_by='RootPrefixAssociation.position')
    prefixes = association_proxy('prefix_assocs', 'prefix')


class ModifiedRoot(Root):

    """A root with one or more modifications."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': 2}

    mod_assocs = relationship('RootModAssociation',
                              collection_class=ordering_list('position'),
                              order_by='RootModAssociation.position')
    modifications = association_proxy('mod_assocs', 'modification')


class PrefixedModifiedRoot(PrefixedRoot, ModifiedRoot):

    """A prefixed root with one or more modifications."""

    __mapper_args__ = {'polymorphic_identity': 3}

    # `mod_assocs` and `modifications` are not defined automatically, despite
    # the diamond inheritance for this class.
    mod_assocs = relationship('RootModAssociation',
                              collection_class=ordering_list('position'),
                              order_by='RootModAssociation.position')
    modifications = association_proxy('mod_assocs', 'modification')


class Stem(SimpleBase):

    """A nominal stem. This corresponds to Panini's **aṅga**:

        | 1.4.13 Anything to which a suffix can be added is called an `aṅga`.

    But although "aṅga" also includes "dhātu," :class:`Stem` does
    not. Verb roots are stored in :class:`Root`.
    """

    genders_id = Column(ForeignKey(GenderGroup.id))
    pos_id = Column(ForeignKey(Tag.id))

    #: ``True`` iff a stem can produce its own words. For stems like "nara"
    #: or "agni" this value is ``True``. For stems like "ja" (dependent on
    #: upapada, as in "agra-ja") or "tva" (a suffix, as in "sama-tva"),
    #: this value is ``False``.
    dependent = Column(Boolean, default=False)

    genders = relationship(GenderGroup)
    pos = relationship(Tag)
    __mapper_args__ = {'polymorphic_on': pos_id}


class NominalStem(Stem):

    """Stem of a :class:`Nominal`."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.NOMINAL}


class PronounStem(Stem):

    """Stem of a :class:`Pronoun`."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.PRONOUN}


class ParticipleStem(Stem):

    """Stem of a :class:`Participle`."""

    __tablename__ = 'participlestem'
    __mapper_args__ = {'polymorphic_identity': Tag.PARTICIPLE}

    id = Column(ForeignKey(Stem.id), primary_key=True)

    root_id = Column(ForeignKey(Root.id))
    mode_id = Column(ForeignKey(Mode.id))
    voice_id = Column(ForeignKey(Voice.id))

    root = relationship(Root, backref='participle_stems')
    mode = relationship(Mode)
    voice = relationship(Voice)


class StemIrregularity(Base):

    """Record of an irregular stem.

    Some Sanskrit stems are inflected irregularly. Since only an
    exceedingly small number of stems is irregular (< 100), it's
    easiest to record those irregularities in their own scheme.
    """

    __tablename__ = 'stem_irregularity'

    id = Column(ForeignKey(Stem.id), primary_key=True)

    #: If ``True``, assume that only the forms stored in the database
    #: are valid. If ``False``, assume that all unspecified forms can
    #: be generated by applying normal rules and endings to the stem.
    fully_described = Column(Boolean)

    stem = relationship(Stem)


# Completed forms
# ---------------
# Nouns, verbs, indeclinables, adjectives, and the like. In Paninian terms,
# these are all `pada`s.

class Form(SimpleBase):

    """A complete form. This corresponds to Panini's **pada**:

        | 1.4.14 Terms ending in "sup" and "tiṅ" are called `pada`.

    In other words, a :class:`Form` is a self-contained linguistic unit that
    could be used in a sentence as-is.
    """

    pos_id = Column(ForeignKey(Tag.id))

    pos = relationship(Tag)
    __mapper_args__ = {'polymorphic_on': pos_id}


class Indeclinable(Form):

    """A complete form. This corresponds to Panini's **avyaya**."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.INDECLINABLE}


class Verb(Form):

    """A complete form. This corresponds to Panini's **tiṅanta**."""

    __tablename__ = 'verb'
    __mapper_args__ = {'polymorphic_identity': Tag.VERB}

    id = Column(ForeignKey(Form.id), primary_key=True)

    root_id = Column(ForeignKey(Root.id))
    vclass_id = Column(ForeignKey(VClass.id))
    person_id = Column(ForeignKey(Person.id))
    number_id = Column(ForeignKey(Number.id))
    mode_id = Column(ForeignKey(Mode.id))
    voice_id = Column(ForeignKey(Voice.id))

    root = relationship(Root, backref='verbs')
    person = relationship(Person)
    number = relationship(Number)
    mode = relationship(Mode)
    voice = relationship(Voice)


class VerbalIndeclinable(Form):

    """A complete form. :class:`VerbalIndeclinable` is a superclass for
    three more specific classes: :class:`Gerund`, :class:`Infinitive`, and
    :class:`PerfectIndeclinable`.
    """

    __tablename__ = 'verbalindeclinable'
    __mapper_args__ = {'polymorphic_identity': Tag.VERBAL_INDECLINABLE}

    id = Column(ForeignKey(Form.id), primary_key=True)
    root_id = Column(ForeignKey(Root.id))

    root = relationship(Root)


class Infinitive(VerbalIndeclinable):

    """A complete form. This corresponds to Panini's **tumanta**."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.INFINITIVE}


class Gerund(VerbalIndeclinable):

    """A complete form. This corresponds to Panini's **ktvānta** and
    **lyabanta**."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.GERUND}


class PerfectIndeclinable(VerbalIndeclinable):

    """A complete form. This corresponds to forms ending in the suffix **ām**,
    as in "īkṣāṃ cakre".
    """

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.PERFECT_INDECLINABLE}


class AbstractNominal(Form):

    """A complete nominal form. This corresponds to Panini's **subanta**."""

    __tablename__ = 'nominal'
    id = Column(ForeignKey(Form.id), primary_key=True)
    stem_id = Column(ForeignKey(Stem.id))
    gender_id = Column(ForeignKey(Gender.id))
    case_id = Column(ForeignKey(Case.id))
    number_id = Column(ForeignKey(Number.id))
    compounded = Column(Boolean)

    stem = relationship(Stem, backref='forms')
    gender = relationship(Gender)
    case = relationship(Case)
    number = relationship(Number)


class Nominal(AbstractNominal):

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.NOMINAL}


class Participle(AbstractNominal):

    """A complete form. This corresponds to Panini's **niṣṭhā** and **sat**.
    Moreover, it also corresponds to **kvasu**."""

    __tablename__ = None
    __mapper_args__ = {'polymorphic_identity': Tag.PARTICIPLE}

    root = association_proxy('stem', 'root')


# Associations
# ------------
# Code for building various many-to-many relationships

class GenderGroupAssociation(Base):

    __tablename__ = 'gender_group_assocs'

    id = Column(Integer, primary_key=True)
    group_id = Column(ForeignKey(GenderGroup.id))
    gender_id = Column(ForeignKey(Gender.id))

    group = relationship(GenderGroup)
    gender = relationship(Gender)

    def __init__(self, gender_id):
        self.gender_id = gender_id


class Paradigm(SimpleBase):

    """Represents an inflectional paradigm. This associates a root with a
    particular class and voice.
    """

    root_id = Column(ForeignKey(Root.id), index=True)
    vclass_id = Column(ForeignKey(VClass.id))
    voice_id = Column(ForeignKey(Voice.id))
    default = Column(Boolean, default=False)

    root = relationship(Root)
    vclass = relationship(VClass)
    voice = relationship(Voice)


class RootPrefixAssociation(SimpleBase):

    """Associates a prefixed root with a list of prefixes."""

    root_id = Column(ForeignKey(PrefixedRoot.id))
    prefix_id = Column(ForeignKey(VerbPrefix.id))
    position = Column(Integer)

    prefix = relationship(VerbPrefix)

    def __init__(self, prefix):
        self.prefix = prefix


class RootModAssociation(SimpleBase):

    """Associates a modified root with a list of modifications."""

    root_id = Column(ForeignKey(ModifiedRoot.id))
    modification_id = Column(ForeignKey(Modification.id))
    position = Column(Integer)

    modification = relationship(Modification)

    def __init__(self, modification):
        self.modification = modification


# Sandhi rules
# ============

class SandhiType(EnumBase):

    """Rule type. Sandhi rules are usually of three types:

    - *external* rules, which act between words
    - *internal* rules, which act between morphemes
    - *general* rules, which act in any context
    """


class SandhiRule(SimpleBase):

    __tablename__ = 'sandhi'

    first = Column(String)
    second = Column(String)
    result = Column(String)
    rule_type = Column(ForeignKey(SandhiType.id))

    def __repr__(self):
        values = (self.id, self.first, self.second, self.result)
        return 'SandhiRule(%r, %r, %r, %r)' % values



# -*- coding: utf-8 -*-
"""
    sanskrit.context
    ~~~~~~~~~~~~~~~~

    Manages the package context. For details, see
    :class:`~sanskrit.context.Context`.

    :license: MIT
"""
from __future__ import print_function
from past.builtins import execfile
import six
import imp
import os

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from .schema import Base, EnumBase, GenderGroup


class Context(object):

    """The package context. In addition to storing basic config information,
    such as the database URI or paths to various data files, a :class:`Context`
    also constructs a :class:`~sqlalchemy.orm.session.Session` class for
    connecting to the database.

    You can populate a context in several ways. For example, you can pass a
    :class:`dict`::

        context = Context({'DATABASE_URI': 'sqlite:///data.sqlite'})

    or a path to a Python module::

        context = Context('project/config.py')

    If you initialize a context from a module, note that only uppercase
    variables will be stored in the context. This lets you use lowercase
    variables as temporary values.

    Config values are stored internally as a :class:`dict`, so you can always
    just use ordinary :class:`dict` methods::

        context.config['FOO'] = 'baz'

    :param config: an object to read from. If this is a string, treat
                   `config` as a module path and load values from that
                   module. Otherwise, treat `config` as a dictionary.
    """

    def __init__(self, config=None, connect=True):
        #: A :class:`dict` of various settings. By convention, all keys are
        #: uppercase. These are used to create :attr:`engine` and
        #: :attr:`session`.
        self.config = {}

        #: The :class:`~sqlalchemy.engine.Engine` that underlies
        #: the :attr:`session`.
        self.engine = None

        #: A :class:`~sqlalchemy.orm.session.Session` class.
        self.session = None

        if isinstance(config, (six.text_type, six.string_types)):
            filepath = config
            config = imp.new_module('config')
            config.__file__ = filepath
            try:
                execfile(filepath, config.__dict__)
            except IOError as e:
                e.strerror = 'Cannot load config file: %s' % e.strerror
                raise

        try:
            config = config or {}
            for key in config:
                if key.isupper():
                    self.config[key] = config[key]
        except TypeError:
            for key in dir(config):
                if key.isupper():
                    self.config[key] = getattr(config, key)

        def default(name, *args):
            path = os.path.join(self.config['DATA_PATH'], *args)
            self.config.setdefault(name, path)

        default('COMPOUNDED_NOMINAL_ENDINGS', 'nominal-endings-compounded.csv')
        default('ENUMS', 'enums.csv')
        default('GERUNDS', 'gerunds.csv')
        default('INDECLINABLES', 'indeclinables.csv')
        default('INFINITIVES', 'infinitives.csv')
        default('INFLECTED_NOMINAL_ENDINGS', 'nominal-endings-inflected.csv')
        default('IRREGULAR_ADJECTIVES', 'irregular-adjectives.csv')
        default('IRREGULAR_NOUNS', 'irregular-nouns.csv')
        default('MODIFIED_ROOTS', 'modified-roots.csv')
        default('NOMINAL_STEMS', 'nominal-stems.csv')
        default('PARTICIPLE_STEMS', 'participle-stems.csv')
        default('PREFIXED_ROOTS', 'prefixed-roots.csv')
        default('PREFIX_GROUPS', 'prefix-groups.csv')
        default('PRONOUNS', 'pronouns.csv')
        default('SANDHI_RULES', 'sandhi-rules.csv')
        default('UNPREFIXED_ROOTS', 'unprefixed-roots.csv')
        default('VERBAL_INDECLINABLES', 'verbal-indeclinables.csv')
        default('VERBS', 'verbs.csv')
        default('VERB_ENDINGS', 'verb-endings.csv')
        default('VERB_PREFIXES', 'verb-prefixes.csv')
        default('VERB_STEMS', 'verb-stems.csv')

        if connect and 'DATABASE_URI' in self.config:
            self.connect()

    def build(self):
        """Build all data."""
        from sanskrit_util import setup
        setup.run(self)

    def connect(self):
        """Connect to the database."""
        self.engine = create_engine(self.config['DATABASE_URI'])
        self.session = scoped_session(sessionmaker(autocommit=False,
                                                   autoflush=False,
                                                   bind=self.engine))

    def create_all(self):
        """Create tables for every model in `sanskrit_util.schema`."""
        metadata = Base.metadata
        extant = {
            t.name for t in metadata.tables.values() if t.exists(self.engine)}
        metadata.create_all(self.engine)
        for name in metadata.sorted_tables:
            if name not in extant:
                print('  [ c ] {0}'.format(name))

    def drop_all(self):
        """Drop all tables defined in `sanskrit_util.schema`."""
        Base.metadata.drop_all(self.engine)

    def _build_enums(self):
        """Fetch and store enumerated data."""
        self._enum_id = {}
        self._enum_abbr = {}
        self._gender_set = {}

        session = self.session
        for cls in EnumBase.__subclasses__():
            key = cls.__tablename__
            self._enum_id[key] = enum_id = {}
            self._enum_abbr[key] = enum_abbr = {}
            for item in session.query(cls).all():
                enum_id[item.name] = enum_id[item.abbr] = item.id
                enum_abbr[item.id] = enum_abbr[item.name] = item.abbr

        for group in session.query(GenderGroup):
            member_ids = set([x.id for x in group.members])
            self._gender_set[group.id] = member_ids

        session.remove()

    @property
    def enum_id(self):
        """Maps a name or abbreviation to an ID."""
        try:
            return self._enum_id
        except AttributeError:
            self._build_enums()
            return self._enum_id

    @property
    def enum_abbr(self):
        """Maps an ID or name to an abbreviation."""
        try:
            return self._enum_abbr
        except AttributeError:
            self._build_enums()
            return self._enum_abbr

    @property
    def gender_set(self):
        try:
            return self._gender_set
        except AttributeError:
            self._build_enums()
            return self._gender_set



# -*- coding: utf-8 -*-
"""
sanskrit.transliterate.sanscript
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Transliteration functions for Sanskrit. The most important function is
:func:`transliterate`, which is very easy to use::

    output = transliterate(data, IAST, DEVANAGARI)

By default, the module supports the following scripts:

- Bengali_
- Devanagari_
- Gujarati_
- Kannada_
- Malayalam_
- Telugu_

and the following romanizations:

- Harvard-Kyoto_
- IAST_ (also known as Roman Unicode)
- SLP1

Each of these **schemes** is defined in a global dictionary `SCHEMES`, whose
keys are strings::

    devanagari_scheme = SCHEMES['devanagari']

For convenience, we also define a variable for each scheme::

    devanagari_scheme = SCHEMES[DEVANAGARI]

These variables are documented below.

:license: MIT and BSD

.. _Bengali: http://en.wikipedia.org/wiki/Bengali_alphabet
.. _Devanagari: http://en.wikipedia.org/wiki/Devanagari
.. _Gujarati: http://en.wikipedia.org/wiki/Gujarati_alphabet
.. _Kannada: http://en.wikipedia.org/wiki/Kannada_alphabet
.. _Malayalam: http://en.wikipedia.org/wiki/Malayalam_alphabet
.. _Telugu: http://en.wikipedia.org/wiki/Telugu_alphabet

.. _Harvard-Kyoto: http://en.wikipedia.org/wiki/Harvard-Kyoto
.. _IAST: http://en.wikipedia.org/wiki/IAST
"""

from __future__ import unicode_literals, division, absolute_import, print_function
import six

# Brahmic schemes
# ---------------
#: Internal name of Bengali. Bengali ``ba`` and ``va`` are both rendered
#: as `ব`.
from builtins import zip
from builtins import object
BENGALI = 'bengali'

#: Internal name of Devanagari.
DEVANAGARI = 'devanagari'

#: Internal name of Gujarati.
GUJARATI = 'gujarati'

#: Internal name of Gurmukhi.
GURMUKHI = 'gurmukhi'

#: Internal name of Kannada.
KANNADA = 'kannada'

#: Internal name of Malayalam.
MALAYALAM = 'malayalam'

#: Internal name of Oriya.
ORIYA = 'oriya'

#: Internal name of Tamil.
TAMIL = 'tamil'

#: Internal name of Telugu.
TELUGU = 'telugu'

#: Internal name of Telugu.
TELUGU = 'telugu'

# Roman schemes
# -------------
#: Internal name of Harvard-Kyoto.
HK = 'hk'

#: Internal name of IAST.
IAST = 'iast'

#: Internal name of ITRANS
ITRANS = 'itrans'

#: Internal name of ITRANS
KOLKATA = 'kolkata'

#: Internal name of SLP1.
SLP1 = 'slp1'

#: Internal name of Velthuis.
VELTHUIS = 'velthuis'

#: Internal name of WX.
WX = 'wx'

SCHEMES = {}


class Scheme(dict):

    """Represents all of the data associated with a given scheme. In addition
    to storing whether or not a scheme is roman, :class:`Scheme` partitions
    a scheme's characters into important functional groups.

    :class:`Scheme` is just a subclass of :class:`dict`.

    :param data: a :class:`dict` of initial values.
    :param is_roman: `True` if the scheme is a romanization and `False`
                     otherwise.
    """

    def __init__(self, data=None, is_roman=True):
        super(Scheme, self).__init__(data or {})
        self.is_roman = is_roman


class SchemeMap(object):

    """Maps one :class:`Scheme` to another. This class grabs the metadata and
    character data required for :func:`transliterate`.

    :param from_scheme: the source scheme
    :param to_scheme: the destination scheme
    """

    def __init__(self, from_scheme, to_scheme):
        """Create a mapping from `from_scheme` to `to_scheme`."""
        self.marks = {}
        self.virama = {}

        self.vowels = {}
        self.consonants = {}
        self.other = {}
        self.from_roman = from_scheme.is_roman
        self.to_roman = to_scheme.is_roman
        self.longest = max(len(x) for g in from_scheme
                           for x in from_scheme[g])

        for group in from_scheme:
            if group not in to_scheme:
                continue
            sub_map = {k: v for (k, v) in zip(from_scheme[group],
                                              to_scheme[group])}
            if group.endswith('marks'):
                self.marks.update(sub_map)
            elif group == 'virama':
                self.virama = sub_map
            else:
                self.other.update(sub_map)
                if group.endswith('consonants'):
                    self.consonants.update(sub_map)
                elif group.endswith('vowels'):
                    self.vowels.update(sub_map)


def _roman(data, scheme_map, **kw):
    """Transliterate `data` with the given `scheme_map`. This function is used
    when the source scheme is a Roman scheme.

    :param data: the data to transliterate
    :param scheme_map: a dict that maps between characters in the old scheme
                       and characters in the new scheme
    """
    vowels = scheme_map.vowels
    marks = scheme_map.marks
    virama = scheme_map.virama
    consonants = scheme_map.consonants
    other = scheme_map.other
    longest = scheme_map.longest
    to_roman = scheme_map.to_roman

    togglers = kw.pop('togglers', set())
    suspend_on = kw.pop('suspend_on', set())
    suspend_off = kw.pop('suspend_off', set())
    if kw:
        raise TypeError('Unexpected keyword argument %s' % list(kw.keys())[0])

    buf = []
    i = 0
    had_consonant = found = False
    len_data = len(data)
    append = buf.append

    # If true, don't transliterate. The toggle token is discarded.
    toggled = False
    # If true, don't transliterate. The suspend token is retained.
    # `suspended` overrides `toggled`.
    suspended = False

    while i <= len_data:
        # The longest token in the source scheme has length `longest`. Iterate
        # over `data` while taking `longest` characters at a time. If we don`t
        # find the character group in our scheme map, lop off a character and
        # try again.
        #
        # If we've finished reading through `data`, then `token` will be empty
        # and the loop below will be skipped.
        token = data[i:i + longest]

        while token:
            if token in togglers:
                toggled = not toggled
                i += 2  # skip over the token
                found = True  # force the token to fill up again
                break

            if token in suspend_on:
                suspended = True
            elif token in suspend_off:
                suspended = False

            if toggled or suspended:
                token = token[:-1]
                continue

            # Catch the pattern CV, where C is a consonant and V is a vowel.
            # V should be rendered as a vowel mark, a.k.a. a "dependent"
            # vowel. But due to the nature of Brahmic scripts, 'a' is implicit
            # and has no vowel mark. If we see 'a', add nothing.
            if had_consonant and token in vowels:
                mark = marks.get(token, '')
                if mark:
                    append(mark)
                elif to_roman:
                    append(vowels[token])
                found = True

            # Catch any other character, including consonants, punctuation,
            # and regular vowels. Due to the implicit 'a', we must explicitly
            # end any lingering consonants before we can handle the current
            # token.
            elif token in other:
                if had_consonant:
                    append(virama[''])
                append(other[token])
                found = True

            if found:
                had_consonant = token in consonants
                i += len(token)
                break
            else:
                token = token[:-1]

        # We've exhausted the token; this must be some other character. Due to
        # the implicit 'a', we must explicitly end any lingering consonants
        # before we can handle the current token.
        if not found:
            if had_consonant:
                append(virama[''])
            if i < len_data:
                append(data[i])
                had_consonant = False
            i += 1

        found = False

    return ''.join(buf)


def _brahmic(data, scheme_map, **kw):
    """Transliterate `data` with the given `scheme_map`. This function is used
    when the source scheme is a Brahmic scheme.

    :param data: the data to transliterate
    :param scheme_map: a dict that maps between characters in the old scheme
                       and characters in the new scheme
    """
    marks = scheme_map.marks
    virama = scheme_map.virama
    consonants = scheme_map.consonants
    other = scheme_map.other
    to_roman = scheme_map.to_roman

    buf = []
    had_consonant = False
    append = buf.append

    for L in data:
        if L in marks:
            append(marks[L])
        elif L in virama:
            append(virama[L])
        else:
            if had_consonant:
                append('a')
            append(other.get(L, L))
        had_consonant = to_roman and L in consonants

    if had_consonant:
        append('a')
    return ''.join(buf)


def transliterate(data, _from=None, _to=None, scheme_map=None, **kw):
    """Transliterate `data` with the given parameters::

        output = transliterate('idam adbhutam', HK, DEVANAGARI)

    Each time the function is called, a new :class:`SchemeMap` is created
    to map the input scheme to the output scheme. This operation is fast
    enough for most use cases. But for higher performance, you can pass a
    pre-computed :class:`SchemeMap` instead::

        scheme_map = SchemeMap(SCHEMES[HK], SCHEMES[DEVANAGARI])
        output = transliterate('idam adbhutam', scheme_map=scheme_map)

    :param data: the data to transliterate
    :param _from: the name of a source scheme
    :param _to: the name of a destination scheme
    :param scheme_map: the :class:`SchemeMap` to use. If specified, ignore
                       `_from` and `_to`. If unspecified, create a
                       :class:`SchemeMap` from `_from` to `_to`.
    """
    if scheme_map is None:
        from_scheme = SCHEMES[_from]
        to_scheme = SCHEMES[_to]
        scheme_map = SchemeMap(from_scheme, to_scheme)

    options = {
        'togglers': set(['##']),
        'suspend_on': set('<'),
        'suspend_off': set('>')
    }
    options.update(kw)

    func = _roman if scheme_map.from_roman else _brahmic
    return func(data, scheme_map, **options)


def _setup():
    """Add a variety of default schemes."""
    s = six.text_type.split

    SCHEMES.update({
        BENGALI: Scheme({
            'vowels': s("""অ আ ই ঈ উ ঊ ঋ ৠ ঌ ৡ এ ঐ ও ঔ"""),
            'marks': s("""া ি ী ু ূ ৃ ৄ ৢ ৣ ে ৈ ো ৌ"""),
            'virama': s('্'),
            'other': s('ং ঃ ঁ'),
            'consonants': s("""
                            ক খ গ ঘ ঙ
                            চ ছ জ ঝ ঞ
                            ট ঠ ড ঢ ণ
                            ত থ দ ধ ন
                            প ফ ব ভ ম
                            য র ল ব
                            শ ষ স হ
                            ळ ক্ষ জ্ঞ
                            """),
            'symbols': s("""
                       ॐ ঽ । ॥
                       ০ ১ ২ ৩ ৪ ৫ ৬ ৭ ৮ ৯
                       """)
        }, is_roman=False),
        DEVANAGARI: Scheme({
            'vowels': s("""अ आ इ ई उ ऊ ऋ ॠ ऌ ॡ ए ऐ ओ औ"""),
            'marks': s("""ा ि ी ु ू ृ ॄ ॢ ॣ े ै ो ौ"""),
            'virama': s('्'),
            'other': s('ं ः ँ'),
            'consonants': s("""
                            क ख ग घ ङ
                            च छ ज झ ञ
                            ट ठ ड ढ ण
                            त थ द ध न
                            प फ ब भ म
                            य र ल व
                            श ष स ह
                            ळ क्ष ज्ञ
                            """),
            'symbols': s("""
                       ॐ ऽ । ॥
                       ० १ २ ३ ४ ५ ६ ७ ८ ९
                       """)
        }, is_roman=False),
        GUJARATI: Scheme({
            'vowels': s("""અ આ ઇ ઈ ઉ ઊ ઋ ૠ ઌ ૡ એ ઐ ઓ ઔ"""),
            'marks': s("""ા િ ી ુ ૂ ૃ ૄ ૢ ૣ ે ૈ ો ૌ"""),
            'virama': s('્'),
            'other': s('ં ઃ ઁ'),
            'consonants': s("""
                            ક ખ ગ ઘ ઙ
                            ચ છ જ ઝ ઞ
                            ટ ઠ ડ ઢ ણ
                            ત થ દ ધ ન
                            પ ફ બ ભ મ
                            ય ર લ વ
                            શ ષ સ હ
                            ળ ક્ષ જ્ઞ
                            """),
            'symbols': s("""
                       ૐ ઽ ૤ ૥
                       ૦ ૧ ૨ ૩ ૪ ૫ ૬ ૭ ૮ ૯
                       """)
        }, is_roman=False),
        GURMUKHI: Scheme({
            'vowels': s("""ਅ ਆ ਇ ਈ ਉ ਊ ऋ ॠ ऌ ॡ ਏ ਐ ਓ ਔ"""),
            'marks': ['ਾ', 'ਿ', 'ੀ', 'ੁ', 'ੂ', '', '',
                      '', '', 'ੇ', 'ੈ', 'ੋ', 'ੌ'],
            'virama': s('੍'),
            'other': s('ਂ ਃ ਁ'),
            'consonants': s("""
                            ਕ ਖ ਗ ਘ ਙ
                            ਚ ਛ ਜ ਝ ਞ
                            ਟ ਠ ਡ ਢ ਣ
                            ਤ ਥ ਦ ਧ ਨ
                            ਪ ਫ ਬ ਭ ਮ
                            ਯ ਰ ਲ ਵ
                            ਸ਼ ਸ਼ ਸ ਹ
                            ਲ਼ ਕ੍ਸ਼ ਜ੍ਞ
                            """),
            'symbols': s("""
                       ॐ ऽ । ॥
                       ੦ ੧ ੨ ੩ ੪ ੫ ੬ ੭ ੮ ੯
                       """)
        }, is_roman=False),
        HK: Scheme({
            'vowels': s("""a A i I u U R RR lR lRR e ai o au"""),
            'marks': s("""A i I u U R RR lR lRR e ai o au"""),
            'virama': [''],
            'other': s('M H ~'),
            'consonants': s("""
                            k kh g gh G
                            c ch j jh J
                            T Th D Dh N
                            t th d dh n
                            p ph b bh m
                            y r l v
                            z S s h
                            L kS jJ
                            """),
            'symbols': s("""
                       OM ' | ||
                       0 1 2 3 4 5 6 7 8 9
                       """)
        }),
        IAST: Scheme({
            'vowels': s("""a ā i ī u ū ṛ ṝ ḷ ḹ e ai o au"""),
            'marks': s("""ā i ī u ū ṛ ṝ ḷ ḹ e ai o au"""),
            'virama': [''],
            'other': s('ṃ ḥ m̐'),
            'consonants': s("""
                            k kh g gh ṅ
                            c ch j jh ñ
                            ṭ ṭh ḍ ḍh ṇ
                            t th d dh n
                            p ph b bh m
                            y r l v
                            ś ṣ s h
                            ḻ kṣ jñ
                            """),
            'symbols': s("""
                       oṃ ' । ॥
                       0 1 2 3 4 5 6 7 8 9
                       """)
        }),
        KANNADA: Scheme({
            'vowels': s("""ಅ ಆ ಇ ಈ ಉ ಊ ಋ ೠ ಌ ೡ ಏ ಐ ಓ ಔ"""),
            'marks': s("""ಾ ಿ ೀ ು ೂ ೃ ೄ ೢ ೣ ೇ ೈ ೋ ೌ"""),
            'virama': s('್'),
            'other': s('ಂ ಃ ँ'),
            'consonants': s("""
                            ಕ ಖ ಗ ಘ ಙ
                            ಚ ಛ ಜ ಝ ಞ
                            ಟ ಠ ಡ ಢ ಣ
                            ತ ಥ ದ ಧ ನ
                            ಪ ಫ ಬ ಭ ಮ
                            ಯ ರ ಲ ವ
                            ಶ ಷ ಸ ಹ
                            ಳ ಕ್ಷ ಜ್ಞ
                            """),
            'symbols': s("""
                       ಓಂ ऽ । ॥
                       ೦ ೧ ೨ ೩ ೪ ೫ ೬ ೭ ೮ ೯
                       """)
        }, is_roman=False),
        MALAYALAM: Scheme({
            'vowels': s("""അ ആ ഇ ഈ ഉ ഊ ഋ ൠ ഌ ൡ ഏ ഐ ഓ ഔ"""),
            'marks': s("""ാ ി ീ ു ൂ ൃ ൄ ൢ ൣ േ ൈ ോ ൌ"""),
            'virama': s('്'),
            'other': s('ം ഃ ँ'),
            'consonants': s("""
                            ക ഖ ഗ ഘ ങ
                            ച ഛ ജ ഝ ഞ
                            ട ഠ ഡ ഢ ണ
                            ത ഥ ദ ധ ന
                            പ ഫ ബ ഭ മ
                            യ ര ല വ
                            ശ ഷ സ ഹ
                            ള ക്ഷ ജ്ഞ
                            """),
            'symbols': s("""
                       ഓം ഽ । ॥
                       ൦ ൧ ൨ ൩ ൪ ൫ ൬ ൭ ൮ ൯
                       """)
        }, is_roman=False),
        ORIYA: Scheme({
            'vowels': s("""ଅ ଆ ଇ ଈ ଉ ଊ ଋ ୠ ଌ ୡ ଏ ଐ ଓ ଔ"""),
            'marks': ['ା', 'ି', 'ୀ', 'ୁ', 'ୂ', 'ୃ', 'ୄ',
                      '', '', 'େ', 'ୈ', 'ୋ', 'ୌ'],
            'virama': s('୍'),
            'other': s('ଂ ଃ ଁ'),
            'consonants': s("""
                            କ ଖ ଗ ଘ ଙ
                            ଚ ଛ ଜ ଝ ଞ
                            ଟ ଠ ଡ ଢ ଣ
                            ତ ଥ ଦ ଧ ନ
                            ପ ଫ ବ ଭ ମ
                            ଯ ର ଲ ଵ
                            ଶ ଷ ସ ହ
                            ଳ କ୍ଷ ଜ୍ଞ
                            """),
            'symbols': s("""
                       ଓଂ ଽ । ॥
                       ୦ ୧ ୨ ୩ ୪ ୫ ୬ ୭ ୮ ୯
                       """)
        }, is_roman=False),
        SLP1: Scheme({
            'vowels': s("""a A i I u U f F x X e E o O"""),
            'marks': s("""A i I u U f F x X e E o O"""),
            'virama': [''],
            'other': s('M H ~'),
            'consonants': s("""
                            k K g G N
                            c C j J Y
                            w W q Q R
                            t T d D n
                            p P b B m
                            y r l v
                            S z s h
                            L kz jY
                            """),
            'symbols': s("""
                       oM ' . ..
                       0 1 2 3 4 5 6 7 8 9
                       """)
        }),
        TAMIL: Scheme({
            'vowels': s("""அ ஆ இ ஈ உ ஊ ऋ ॠ ऌ ॡ ஏ ஐ ஓ ஔ"""),
            'marks': ['ா', 'ி', 'ீ', 'ு', 'ூ', '', '',
                      '', '', 'ே', 'ை', 'ோ', 'ௌ'],
            'virama': s('்'),
            'other': s('ஂ ஃ ँ'),
            'consonants': s("""
                            க க க க ங
                            ச ச ஜ ச ஞ
                            ட ட ட ட ண
                            த த த த ந
                            ப ப ப ப ம
                            ய ர ல வ
                            ஶ ஷ ஸ ஹ
                            ள க்ஷ ஜ்ஞ
                            """),
            'symbols': s("""
                       ௐ ऽ । ॥
                       ௦ ௧ ௨ ௩ ௪ ௫ ௬ ௭ ௮ ௯
                       """)
        }, is_roman=False),
        TELUGU: Scheme({
            'vowels': s("""అ ఆ ఇ ఈ ఉ ఊ ఋ ౠ ఌ ౡ ఏ ఐ ఓ ఔ"""),
            'marks': s("""ా ి ీ ు ూ ృ ౄ ౢ ౣ ే ై ో ౌ"""),
            'virama': s('్'),
            'other': s('ం ః ఁ'),
            'consonants': s("""
                            క ఖ గ ఘ ఙ
                            చ ఛ జ ఝ ఞ
                            ట ఠ డ ఢ ణ
                            త థ ద ధ న
                            ప ఫ బ భ మ
                            య ర ల వ
                            శ ష స హ
                            ళ క్ష జ్ఞ
                            """),
            'symbols': s("""
                       ఓం ఽ । ॥
                       ౦ ౧ ౨ ౩ ౪ ౫ ౬ ౭ ౮ ౯
                       """)
        }, is_roman=False)
    })

_setup()



import collections


class HashTrie(object):

    """A fast trie, for short items."""

    def __init__(self):
        self.mapper = collections.defaultdict(set)
        self.len_longest = 0
        self.lengths = range(1, self.len_longest + 1)

    def __getitem__(self, key):
        return set.union(*[self.mapper[key[:i]] for i in self.lengths])

    def __setitem__(self, key, value):
        self.mapper[key].add(value)
        self.len_longest = max(len(key), self.len_longest)
        self.lengths = range(1, self.len_longest + 1)



# -*- coding: utf-8 -*-
"""
    sanskrit_util.sandhi
    ~~~~~~~~~~~~~~~

    Classes that apply and undo sandhi rules.

    :license: MIT
"""

from builtins import next
from builtins import zip
from builtins import range
from builtins import object
import six
from . import sounds
from .util import HashTrie


class Exempt(six.text_type):

    """A helper class for marking strings as exempt from sandhi changes. To
    mark a string as exempt, just do the following::

        original = 'amI'
        exempt = Exempt('amI')

    :class:`Exempt` is a subclass of :class:`unicode`, so you can use normal
    string methods on :class:`Exempt` objects.
    """


class SandhiObject(object):

    def add_rules(self, rules):
        raise NotImplementedError


class Joiner(SandhiObject):

    """Joins multiple Sanskrit terms by applying sandhi rules."""

    def __init__(self, rules=None):
        self.data = {}
        if rules:
            self.add_rules(rules)

    def add_rules(self, rules):
        """Add rules for joining words.

        Example usage::

            joiner.add_rules[('a', 'i', 'e'), ('a', 'a', 'A'])

        :param rules: a list of 3-tuples, each of which contains:

        - the first part of the combination
        - the second part of the combination
        - the result

        """
        self.data = {}
        for first, second, result in rules:
            self.data[(first, second)] = result

    @staticmethod
    def internal_retroflex(term):
        """Apply the "n -> ṇ" and "s -> ṣ" rules of internal sandhi.

        :param term: the string to process
        """
        # causes "s" retroflexion
        s_trigger = set('iIuUfFeEoOkr')
        # causes "n" retroflexion
        n_trigger = set('fFrz')
        # Allowed after n_trigger
        n_between = sounds.VOWELS.union('kKgGNpPbBmhvyM')
        # Must appear after the retroflexed "n"
        n_after = sounds.VOWELS.union('myvn')
        # Defines t retroflexion
        retroflexion_dict = dict(zip('tT', 'wW'))

        letters = list(term)

        apply_s = False
        apply_n = False
        had_n = False  # Used for double retroflexion ('nisanna' -> 'nizaRRa')
        had_s = False  # Used for 't' retroflexion
        for i, L in enumerate(letters[:-1]):
            # "t" retroflexion after "s" retroflexion
            if had_s:
                had_s = False
                letters[i] = retroflexion_dict.get(L, L)

            # "s" retroflexion
            if apply_s and L == 's':
                letters[i] = L = 'z'
                had_s = True
            apply_s = L in s_trigger

            # "n" retroflexion
            if had_n and L == 'n':
                letters[i] = 'R'
                had_n = False
            elif apply_n and L == 'n' and letters[i + 1] in n_after:
                letters[i] = 'R'
                had_n = True
            if L in n_trigger:
                apply_n = True
            else:
                apply_n = apply_n and L in n_between

        return ''.join(letters)

    def join(self, chunks, internal=False):
        """Join the given chunks according to the object's rules::

            assert 'tasyecCA' == s.join('tasya', 'icCA')

        :meth:`join` does **not** take pragṛhya rules into account. As a
        reminder, the main exception are:

            | 1.1.11 "ī", "ū", and "e" when they end words in the dual.
            | 1.1.12 the same vowels after the "m" of adas;
            | 1.1.13 particles with just one vowel, apart from "ā"
            | 1.1.14 particles that end in "o".

        One simple way to account for these rules is to wrap exempt strings
        with :class:`Exempt`::

            assert joiner.join('te', 'iti') == 'ta iti'
            assert joiner.join(Exempt('te'), 'iti') == 'te iti'

        :param chunks: a list of the strings that should be joined
        :param internal: if true, join words using the empty string instead of
                         `' '`.
        """
        separator = '' if internal else ' '

        it = iter(chunks)
        returned = next(it)
        for chunk in it:
            if not chunk:
                continue
            if isinstance(returned, Exempt):
                returned += separator + chunk
            else:
                # `i` controls the number of letters to grab from the end of
                # the first word. For most rules, one letter is sufficient.
                # But visarga sandhi needs slightly more context.
                for i in (2, 1, 0):
                    if not i:
                        returned += separator + chunk
                        break
                    key = (returned[-i:], chunk[0])
                    result = self.data.get(key, None)
                    if result:
                        returned = returned[:-i] + result + chunk[1:]
                        break
            if isinstance(chunk, Exempt):
                returned = Exempt(returned)

        if internal:
            return Joiner.internal_retroflex(returned)
        else:
            return returned


class Splitter(object):

    """Splits Sanskrit terms by undoing sandhi rules."""

    def __init__(self, rules=None):
        """"""
        self.data = HashTrie()
        if rules:
            self.add_rules(rules)

    def add_rules(self, rules):
        for first, second, result in rules:
            result = result.replace(' ', '')
            items = (first, second, result, len(first), len(second),
                     len(result))
            self.data[result] = items

    def iter_splits(self, chunk):
        """Return a generator for all splits in `chunk`. Results are yielded
        as 2-tuples containing the term before the split and the term after::

            for item in s.splits('nareti'):
                before, after = item

        :meth:`splits` will generate many false positives, usually when the
        first part of the split ends in an invalid consonant::

            assert ('narAv', 'iti') in s.splits('narAviti')

        These should be filtered out in the calling function.

        Splits are generated from left to right, but the function makes no
        guarantees on when certain rules are applied. That is, output is
        loosely ordered but nondeterministic.
        """

        chunk_len = len(chunk)

        for i in range(chunk_len):
            # Default split: chop the chunk in half with no other changes.
            # This can yield a lot of false positives.
            chunk1, chunk2 = chunk[:i], chunk[i:]
            if i:
                yield (chunk1, chunk2)

            # Rule-based splits: undo a sandhi change
            rules = self.data[chunk2]
            for first, second, result, _, _, len_result in rules:
                before = chunk1 + first
                after = second + chunk2[len_result:]
                yield (before, after)

        # Non-split: yield the chunk as-is.
        yield (chunk, '')



"""
    sanskrit_util.shell
    ~~~~~~~~~~~~~~

    A convenient REPL for fixing errors in the tagger output.

    :license: MIT
"""
from __future__ import print_function
from builtins import input
from builtins import range
from builtins import object
import re

from sanskrit_util import tagger


class Color(object):

    """Terminal colors. Unix systems only."""

    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'

    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    END = '\033[0m'


def with_color(s, colors):
    """Convenience function that adds `Color.END`."""
    return colors + s + Color.END


def print_with_highlighted_chunk(segment, n):
    """Returns a clean version of `segment` with chunk `n` in red."""
    chunks = [x.split() for x in segment.strip().splitlines()]
    i = 0
    for L, line in enumerate(chunks):
        for c, chunk in enumerate(line):
            if i == n:
                s = chunks[L][c]
                chunks[L][c] = with_color(s, Color.RED + Color.BOLD)
                print('\n'.join([' '.join(x) for x in chunks]))
                return
            else:
                i += 1


def print_numbered_items(items, ctx):
    """Convenience function to print a numbered list of items."""
    for i, item in enumerate(items):
        hrf = item.human_readable_form(ctx)
        pos = hrf[1]
        print('{} : {}'.format((i + 1), hrf))


def int_or_none(x, limit):
    """Returns `int(x)` if `x` is a valid `int` or `None` otherwise.

    `x` is valid if `1 <= x <= limit`.
    """
    try:
        value = int(x)
        if 1 <= value <= limit:
            return value
        else:
            return None
    except ValueError:
        return None


def print_help():
    print(with_color('1  : get alternatives for form 1', Color.GREEN))
    print(with_color('s  : re-split the chunk', Color.GREEN))
    print(with_color('pc : previous chunk', Color.GREEN))
    print(with_color('nc : next chunk', Color.GREEN))
    print(with_color('q  : quit', Color.GREEN))  # TODO: save and quit
    print(with_color('?  : help', Color.GREEN))


def segment_repl(ctx, segment):
    print(with_color('*' + '-' * 50, Color.GREEN + Color.BOLD))
    print(with_color('*', Color.GREEN + Color.BOLD))
    print(with_color('* Interactive Sanskrit tagger', Color.GREEN + Color.BOLD))
    print(with_color('*', Color.GREEN + Color.BOLD))
    print(with_color('*' + '-' * 50, Color.GREEN + Color.BOLD))
    print_help()

    t = tagger.Tagger(ctx)

    # Chunk index -> list of TaggedItems
    items_for_chunk = {}
    for item in t.tag(segment):
        items_for_chunk.setdefault(item.chunk_index, []).append(item)

    # Chunk index -> highlighted chunk
    index = 0
    chunks = list(t.iter_chunks(segment))
    n = len(chunks)

    while index < n:
        if re.match("^[0-9|./]+$", chunks[index]):
            index += 1
            continue

        print(with_color('~' * 50, Color.BLUE + Color.BOLD))
        print()
        print_with_highlighted_chunk(segment, index)
        print()
        print_numbered_items(items_for_chunk[index], ctx)
        print()

        command = input(": ")
        int_command = int_or_none(command, limit=len(items_for_chunk[index]))

        # All OK
        if not command or command == 'nc':
            index += 1

        elif command == '?' or command == 'help':
            print_help()

        # Go back one
        elif command == 'pc':
            if index > 0:
                index -= 1

        elif command == 'q':
            return

        # Bad split
        elif command == 's':
            split = input("Split: ")
            items_for_chunk[index] = list(t.tag_segment(split))

        # Bad form
        elif int_command is not None:
            bad_item = items_for_chunk[index][int_command - 1]
            candidate_items = []
            for form in t.analyzer.analyze(bad_item.form.name):
                candidate_items.append(
                    tagger.TaggedItem(bad_item.segment_id,
                                      bad_item.chunk_index, form))

            print_numbered_items(candidate_items, ctx)

            choice = None
            while choice is None:
                choice = int_or_none(input('Which? : '),
                                     limit=len(candidate_items))

            items_for_chunk[index][int_command - 1] = candidate_items[choice - 1]

        else:
            print()
            msg = "Unrecognized command '%s'. Try one of these:" % command
            print(with_color(msg, Color.BOLD + Color.GREEN))
            print_help()

    # All done!
    # TODO: write to file
    for i in range(n):
        for item in items_for_chunk[i]:
            print(item.human_readable_form(ctx))


def run(ctx, segment):
    segment_repl(ctx, segment)



"""
    sanskrit_util.tagger
    ~~~~~~~~~~~~~~~
    Code for converting Sanskrit paragraphs into a list of linguistic
    forms. This is done through a *part-of-speech tagger*
    (:class:`~sanskrit_util.tagger.Tagger`) that also removes sandhi and
    identifies the lexical roots that underlie the forms in some passage.

    When tagging, a block of text (i.e. a verse or paragraph) is called
    a *segment* and its space-separated substrings are called *chunks*.
    For example, the segment ``'aTa SabdAnuSAsanam |'`` has chunks
    ``'aTa'``, ``'SabdAnuSAsanam'``, and ``'|'``.

    :license: MIT
"""

from sanskrit_util import analyze, models, sandhi, schema, util


class NonForm(object):

    """Wraps a chunk that couldn't be parsed by the :class:`Tagger`."""

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "NonForm('{}')".format(self.name)


class TaggedItem(object):

    """Associates a linguistic form with a specific chunk and segment."""

    def __init__(self, segment_id, chunk_index, form):
        self.segment_id = segment_id
        self.chunk_index = chunk_index
        self.form = form

    def __repr__(self):
        fields = (self.segment_id, self.chunk_index, self.form.name)
        return 'TaggedItem({})'.format(fields)

    def _enum_string(self, ctx, fields):
        strings = []
        enums = ctx.enum_abbr
        for field in fields:
            id = getattr(self.form, field + '_id')
            if id is not None:
                strings.append(enums[field][id])
            else:
                strings.append('')
        return '-'.join(strings)

    def tag(self, ctx):
        form = self.form
        if isinstance(form, NonForm):
            return models.SEQUENCE_BOUNDARY
        if isinstance(form, schema.Indeclinable):
            tup = ('indeclinable',)
        elif isinstance(form, schema.Verb):
            tup = ('verb', self._enum_string(ctx, ['person', 'number']))
        elif isinstance(form, schema.Nominal):
            tup = ('nominal',
                    self._enum_string(ctx, ['gender', 'case', 'number']))
        elif isinstance(form, schema.Infinitive):
            tup = ('infinitive',)
        elif isinstance(form, schema.Gerund):
            tup = ('gerund',)
        elif isinstance(form, schema.PerfectIndeclinable):
            tup = ('perfect-indeclinable',)
        return '-'.join(tup)

    def human_readable_form(self, ctx):
        form = self.form
        if isinstance(form, NonForm):
            return (form.name, '', '', '')
        elif isinstance(form, schema.Indeclinable):
            return (form.name, 'indeclinable', '', '')
        elif isinstance(form, schema.Verb):
            return (form.name, 'verb', form.root.name,
                     self._enum_string(ctx,
                                       ['vclass', 'person', 'number', 'mode',
                                        'voice']))
        elif isinstance(form, schema.Nominal):
            return (form.name, 'nominal', form.stem.name,
                     self._enum_string(ctx, ['gender', 'case', 'number']))
        elif isinstance(form, schema.Infinitive):
            return (form.name, 'infinitive', form.root.name, '')
        elif isinstance(form, schema.Gerund):
            return (form.name, 'gerund', form.root.name, '')
        elif isinstance(form, schema.PerfectIndeclinable):
            return (form.name, 'perfect-indeclinable', form.root.name, '')


class Tagger(object):

    """The part-of-speech tagger."""

    def __init__(self, ctx):
        rules = [(x.first, x.second, x.result)
                 for x in ctx.session.query(schema.SandhiRule).all()]

        self.ctx = ctx
        self.splitter = sandhi.Splitter(rules)
        self.analyzer = analyze.SimpleAnalyzer(ctx)
        self.model = models.FeatureModel()

    def _score(self, before, cur, remainder):
        """Compute a score over the given tagger state."""
        # xs = [before[-1].tag(self.ctx)] if before else []
        # y = cur.tag(self.ctx)
        # return self.model.log_cond_prob(xs, y)
        return self.model.score(cur, remainder)

    def iter_chunks(self, segment):
        """Iterate over the chunks in `segment`.

        :param segment: an arbitrary string
        """
        for line in segment.splitlines():
            for chunk in line.split():
                yield chunk

    def tag(self, segment, segment_id=None):
        """Return the linguistic forms that compose `segment`. If a form
        can't be parsed, it's wrapped in :class:`NonForm`.

        :param segment: an arbitrary string
        :return: a list of :class:`TaggedItem` objects.
        """
        chunks = list(self.iter_chunks(segment))
        if not chunks:
            return

        q = util.PriorityQueue()
        q.push(([], 0, chunks[0]), 0)

        done = []
        while q:
            (done, chunk_index, remainder), priority = q.pop_with_priority()
            # Chunk is done
            if not remainder:
                if chunk_index + 1 < len(chunks):
                    new_state = (done, chunk_index + 1, chunks[chunk_index + 1])
                    q.push(new_state, priority)
                    continue
                else:
                    # Segment is done!
                    break

            for before, after in self.splitter.iter_splits(remainder):
                # Without this line, the tagger could loop forever. This
                # looping occurs if a sandhi rule has the form "X -> Y X",
                # which yields Y while leaving the term with X unchanged.
                if remainder == after: continue

                results = self.analyzer.analyze(before)
                for result in results:
                    item = TaggedItem(segment_id, chunk_index, result)
                    q.push((done + [item], chunk_index, after),
                            priority + self._score(done, item, after))

            # Add "default" state in case nothing could be found.
            if remainder == chunks[chunk_index]:
                result = NonForm(remainder)
                item = TaggedItem(segment_id, chunk_index, result)
                new_state = (done + [item], chunk_index, None)
                q.push(new_state, priority + self._score(done, item, remainder))

        return done



"""
sanskrit.setup
~~~~~~~~~~~~~~

Setup code for various Sanskrit data.
"""
from __future__ import print_function

import sys

from sanskrit_util import util
from sanskrit_util.schema import *

import sqlalchemy.schema

# Populated in `add_enums`
ENUM = {}


# Miscellaneous
# -------------

def add_tags(ctx):
    """Populate the `Tag` table."""
    session = ctx.session
    for key in dir(Tag):
        if key.isupper():
            id = getattr(Tag, key)
            session.add(Tag(id=id, name=key.lower()))
    session.commit()
    session.close()


def add_enums(ctx):
    """Add enumerated data to the database. Among others, this includes:

    - persons
    - numbers
    - modes
    - voices
    - genders
    - cases

    and any other data with small, known limits.
    """

    session = ctx.session
    type_to_class = {
        'case': Case,
        'class': VClass,
        'gender': Gender,
        'gender_group': GenderGroup,
        'modification': Modification,
        'mode': Mode,
        'number': Number,
        'person': Person,
        'sandhi_rule_type': SandhiType,
        'voice': Voice,
    }

    # First pass: ordinary enums
    for row in util.read_csv(ctx.config['ENUMS']):
        if row['enum_type'] == 'gender_group':
            continue

        cls = type_to_class.get(row['enum_type'], None)
        # TODO: always non-None?
        if cls is None:
            continue

        enum_abbr = cls.__tablename__
        if enum_abbr not in ENUM:
            util.tick(cls.__name__)
        ENUM.setdefault(enum_abbr, {})

        abbreviation = row['abbreviation']
        e = cls(name=row['human_readable_value'], abbr=abbreviation)
        session.add(e)
        session.flush()
        ENUM[enum_abbr][abbreviation] = e.id
    session.commit()

    # Second pass: gender groups
    for row in util.read_csv(ctx.config['ENUMS']):
        if row['enum_type'] != 'gender_group':
            continue

        cls = type_to_class.get(row['enum_type'], None)
        enum_abbr = cls.__tablename__
        if enum_abbr not in ENUM:
            util.tick(cls.__name__)
        ENUM.setdefault(enum_abbr, {})

        abbreviation = row['abbreviation']
        e = cls(name=row['human_readable_value'], abbr=abbreviation)
        session.add(e)
        session.flush()

        if set(abbreviation).issubset('mfn'):
            e.members = [ENUM['gender'][x] for x in abbreviation]

        ENUM[enum_abbr][abbreviation] = e.id

    session.commit()
    session.close()


def add_sandhi_rules(ctx):
    """Add sandhi rules to the database."""
    session = ctx.session
    stype = ENUM['sandhi_type']

    for row in util.read_csv(ctx.config['SANDHI_RULES']):
        session.add(SandhiRule(first=row['first'], second=row['second'],
                               result=row['result'],
                               rule_type=stype[row['type']]))
    session.commit()
    session.close()


def add_indeclinables(ctx):
    """Add indeclinables to the database."""
    session = ctx.session
    tick = util.tick_every(300)

    for row in util.read_csv(ctx.config['INDECLINABLES']):
        session.add(Indeclinable(name=row['name']))
        tick(row['name'])

    session.commit()
    session.close()


def add_verb_prefixes(ctx):
    """Add verb prefixes to the database."""
    session = ctx.session
    prefix_map = {}

    for row in util.read_csv(ctx.config['VERB_PREFIXES']):
        # TODO: use prefix type?
        prefix = VerbPrefix(name=row['name'])
        session.add(prefix)
        session.flush()
        prefix_map[row['name']] = prefix.id

    session.commit()
    session.close()
    return prefix_map


def add_verb_endings(ctx):
    """Add verb endings to the database."""
    session = ctx.session
    person = ENUM['person']
    number = ENUM['number']
    mode = ENUM['mode']
    voice = ENUM['voice']

    for row in util.read_csv(ctx.config['VERB_ENDINGS']):
        session.add(VerbEnding(name=row['ending'],
                               category=row['category'],
                               person_id=person[row['person']],
                               number_id=number[row['number']],
                               mode_id=mode[row['mode']],
                               voice_id=voice[row['voice']]))
    session.commit()
    session.close()


def add_roots(ctx, prefix_map):
    """Populates :class:`Root` and its subclasses."""

    # TODO: modified roots
    session = ctx.session
    e_vclass = ENUM['vclass']
    e_voice = ENUM['voice']

    root_map = {}  # (name, hom) -> id

    # First pass: Root
    tick = util.tick_every(100)
    for row in util.read_csv(ctx.config['UNPREFIXED_ROOTS']):
        name, hom = row['root'], row['hom']

        # A root can have multiple paradigms (= multiple appearances)
        if (name, hom) in root_map:
            continue

        root = Root(name=name)
        session.add(root)
        session.flush()
        root_map[(name, hom)] = root.id

        tick(name)

    # Second pass: Paradigm
    for row in util.read_csv(ctx.config['UNPREFIXED_ROOTS']):
        name, hom = row['root'], row['hom']
        vclass, voice = row['class'], row['voice']

        assert (name, hom) in root_map
        root_id = root_map[(name, hom)]
        paradigm = Paradigm(root_id=root_id, vclass_id=e_vclass[vclass],
                            voice_id=e_voice[voice])
        session.add(paradigm)

    session.commit()

    # Prefixed roots
    for i, row in enumerate(util.read_csv(ctx.config['PREFIXED_ROOTS'])):
        name = row['prefixed_root']
        basis = row['unprefixed_root']
        hom = row['hom']
        prefixes = row['prefixes'].split('-')

        assert (basis, hom) in root_map
        basis_id = root_map[(basis, hom)]
        for prefix in prefixes:
            # TODO
            pass

        prefixed_root = PrefixedRoot(name=name, basis_id=basis_id)
        session.add(prefixed_root)
        session.flush()
        root_map[(name, hom)] = prefixed_root.id

        tick(name)

    session.commit()
    session.close()

    return root_map


def add_verbs(ctx, root_map):
    """Add inflected verbs to the database."""

    session = ctx.session
    vclass = ENUM['vclass']
    person = ENUM['person']
    number = ENUM['number']
    mode = ENUM['mode']
    voice = ENUM['voice']
    skipped = set()
    i = 0

    for row in util.read_csv(ctx.config['VERBS']):
        root = row['root']
        hom = row['hom']

        try:
            root_id = root_map[(root, hom)]
        except KeyError:
            skipped.add((root, hom))
            continue

        data = {
            'name': row['form'],
            'root_id': root_id,
            'vclass_id': vclass[row['class']] if row['class'] else None,
            'person_id': person[row['person']],
            'number_id': number[row['number']],
            'mode_id': mode[row['mode']],
            'voice_id': voice[row['voice']]
        }
        session.add(Verb(**data))

        i += 1
        if i % 1000 == 0:
            util.tick(row['form'])
            session.commit()

    session.commit()
    session.close()
    print('Skipped', len(skipped), 'roots.')


def add_verbal_indeclinables(ctx, root_map):
    session = ctx.session
    root_map = root_map or {}
    skipped = set()

    for row in util.read_csv(ctx.config['VERBAL_INDECLINABLES']):
        root, hom, pos = row['root'], row['hom'], row['pos']

        try:
            root_id = root_map[(root, hom)]
        except KeyError:
            skipped.add((root, hom))
            continue

        # TODO: modifications!
        datum = {
            'name': row['form'],
            'root_id': root_id
        }
        if pos == 'gerund':
            session.add(Gerund(**datum))
        elif pos == 'infinitive':
            session.add(Infinitive(**datum))
        else:
            assert False

    session.commit()


def add_participle_stems(ctx, root_map):
    """Populates `ParticipleStem`."""

    session = ctx.session
    root_map = root_map or {}
    mode = ENUM['mode']
    voice = ENUM['voice']
    skipped = set()
    i = 0

    for row in util.read_csv(ctx.config['PARTICIPLE_STEMS']):
        root = row['root']
        hom = row['hom']

        try:
            root_id = root_map[(root, hom)]
        except KeyError:
            skipped.add((root, hom))
            continue

        data = {
            'name': row['stem'],
            'root_id': root_id,
            'mode_id': mode[row['mode']],
            'voice_id': voice[row['voice']]
        }

        session.add(ParticipleStem(**data))

        i += 1
        if i % 100 == 0:
            util.tick(row['stem'])
            session.commit()

    session.commit()
    session.close()
    print('Skipped', len(skipped), 'roots.')


def add_nominal_endings(ctx):
    """Populates `NominalEnding`."""
    session = ctx.session
    gender = ENUM['gender']
    case = ENUM['case']
    number = ENUM['number']

    for row in util.read_csv(ctx.config['COMPOUNDED_NOMINAL_ENDINGS']):
        ending = NominalEnding(name=row['ending'],
                               stem_type=row['stem_type'],
                               gender_id=gender[row['form_gender']],
                               case_id=None, number_id=None, compounded=True)
        session.add(ending)

    seen = set()
    for row in util.read_csv(ctx.config['INFLECTED_NOMINAL_ENDINGS']):
        ending = NominalEnding(name=row['ending'],
                               stem_type=row['stem_type'],
                               gender_id=gender[row['form_gender']],
                               case_id=case[row['case']],
                               number_id=number[row['number']], compounded=False)
        session.add(ending)

        if row['stem_type'] not in seen:
            util.tick(row['stem_type'])
            seen.add(row['stem_type'])

    session.commit()
    session.close()


def add_nominal_stems(ctx):
    """Add regular noun stems to the database."""
    # Since there are so many nominal stems, the SQLAlchemy calls here
    # are a little more low-level.

    conn = ctx.engine.connect()
    ins = NominalStem.__table__.insert()
    gender_group = ENUM['gender_group']

    buf = []
    i = 0
    tick = util.tick_every(5000)
    for row in util.read_csv(ctx.config['NOMINAL_STEMS']):
        genders_id = gender_group[row['stem_genders']]

        buf.append({
            'name': row['stem'],
            'pos_id': Tag.NOMINAL,
            'genders_id': genders_id,
        })

        tick(row['stem'])
        i += 1
        if i % 500 == 0:
            conn.execute(ins, buf)
            buf = []

    # Add any remainder.
    if buf:
        conn.execute(ins, buf)


def add_irregular_nouns(ctx):
    """Add irregular nouns to the database."""

    session = ctx.session
    gender_group = ENUM['gender_group']
    gender = ENUM['gender']
    case = ENUM['case']
    number = ENUM['number']

    with open(ctx.config['IRREGULAR_NOUNS']) as f:
        for noun in yaml.load_all(f):
            genders_id = gender_group[noun['genders']]
            stem = NounStem(name=noun['name'], genders_id=genders_id)
            session.add(stem)
            session.flush()

            # Mark the stem as irregular
            complete = noun['complete']
            irreg = StemIrregularity(stem=stem, fully_described=complete)
            session.add(irreg)
            session.flush()

            util.tick(stem.name)

            for form in noun['forms']:
                name = form['name']
                gender_id = gender[form['gender']]
                case_id = case[form['case']]
                number_id = number[form['number']]

                result = Noun(stem=stem, name=name, gender_id=gender_id,
                              case_id=case_id, number_id=number_id)
                session.add(result)
                session.flush()

    session.commit()
    session.close()


def add_irregular_adjectives(ctx):
    """Add regular irregular adjectives to the database."""

    session = ctx.session
    gender = ENUM['gender']
    case = ENUM['case']
    number = ENUM['number']

    with open(ctx.config['IRREGULAR_ADJECTIVES']) as f:
        for adj in yaml.load_all(f):
            stem = AdjectiveStem(name=adj['name'])
            session.add(stem)
            session.flush()

            # Mark the stem as irregular
            complete = adj['complete']
            irreg = StemIrregularity(stem=stem, fully_described=complete)
            session.add(irreg)
            session.flush()

            util.tick(stem.name)

            for form in adj['forms']:
                name = form['name']
                gender_id = gender[form['gender']]
                case_id = case[form['case']]
                number_id = number[form['number']]

                result = Adjective(stem=stem, name=name, gender_id=gender_id,
                                   case_id=case_id, number_id=number_id)
                session.add(result)

    session.commit()
    session.close()


def add_pronouns(ctx):
    """Populates `PronounStem` and `Pronoun`."""

    session = ctx.session
    gender_group = ENUM['gender_group']
    gender = ENUM['gender']
    case = ENUM['case']
    number = ENUM['number']

    seen_stems = {}  # (stem, genders_id) -> id
    for row in util.read_csv(ctx.config['PRONOUNS']):
        stem = row['stem']
        genders_id = gender_group[row['stem_genders']]

        if (stem, genders_id) not in seen_stems:
            pronoun_stem = PronounStem(name=stem, genders_id=genders_id)
            session.add(pronoun_stem)
            session.flush()
            seen_stems[(stem, genders_id)] = pronoun_stem.id
            util.tick(stem)

        stem_id = seen_stems[(stem, genders_id)]
        session.add(Nominal(stem_id=stem_id, name=row['form'],
                            gender_id=gender[row['form_gender']],
                            case_id=case[row['case']],
                            number_id=number[row['number']]))
        session.flush()

    session.commit()
    session.close()


def run(ctx):
    """Create and populate tables in the database."""
    ctx.drop_all()
    ctx.create_all()

    util.heading('Metadata and sandhi')
    add_tags(ctx)
    add_enums(ctx)
    add_sandhi_rules(ctx)

    util.heading('Indeclinables (non-verbal)')
    add_indeclinables(ctx)

    util.heading('Verbal data')
    prefix_map = add_verb_prefixes(ctx)
    root_map = add_roots(ctx, prefix_map=prefix_map)
    add_verb_endings(ctx)
    add_verbs(ctx, root_map)
    add_participle_stems(ctx, root_map)
    add_verbal_indeclinables(ctx, root_map)
    del prefix_map

    util.heading('Nominal data')
    add_nominal_stems(ctx)
    add_nominal_endings(ctx)
    add_pronouns(ctx)
    # add_irregular_nouns(ctx)
    # add_irregular_adjectives(ctx)

    print('Done.')


if __name__ == '__main__':
    ctx = Context(sys.argv[1])
    run(ctx)



from setuptools import setup, find_packages


setup(
    name='sanskrit_util',
    version='0.1.1.dev1',
    description='Sanskrit utilities',
    long_description='Collection of utilities for analyzing and working with Sanskrit',

    url='https://github.com/sanskrit/sanskrit',
    author='learnsanskrit.org',
    author_email='info@learnsanskrit.org',
    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='sanskrit',

    packages=find_packages(exclude=['docs', 'test*']),
    install_requires=['sqlalchemy >= 0.7', 
                      'future',
                      'six'],
    extras_require={
        'dev': ['pytest'],
        'test': ['pytest'],
    }
)



# -*- coding: utf-8 -*-
"""
    sanskrit_util.sounds
    ~~~~~~~~~~~~~~~

    Code for checking and transforming Sanskrit sounds. This module also
    contains basic metrical functions (see :func:`sanskrit_util.sounds.meter`
    and :func:`sanskrit_util.sounds.num_syllables`).

    All functions assume SLP1.

    :license: MIT
"""

#: All legal sounds, including anusvara, ardhachandra, and Vedic `'L'`.
from builtins import map
from builtins import zip
ALL_SOUNDS = frozenset("aAiIuUfFxXeEoOMHkKgGNcCjJYwWqQRtTdDnpPbBmyrlLvSzsh'~")

#: All legal tokens, including sounds, punctuation (`'|'`), and whitespace.
ALL_TOKENS = ALL_SOUNDS | {'|', ' ', '\n'}

#: All vowels.
VOWELS = frozenset('aAiIuUfFxXeEoO')

#: Short vowels.
SHORT_VOWELS = frozenset('aiufx')

#: Stop consonants.
STOPS = frozenset('kKgGcCjJwWqQtTdDpPbB')

#: Nasals.
NASALS = frozenset('NYRnm')

#: Semivowels.
SEMIVOWELS = frozenset('yrlLv')

#: Savarga
SAVARGA = frozenset('Szsh')

#: Consonants.
CONSONANTS = STOPS.union(NASALS).union(SEMIVOWELS).union(SAVARGA)

#: Valid word-final sounds.
VALID_FINALS = frozenset('aAiIuUfeEoOkwtpNnmsr')


# General functions
# -----------------

def clean(phrase, valid):
    """Remove all characters from `phrase` that are not in `valid`.

    :param phrase: the phrase to clean
    :param valid: the set of valid characters. A sensible default is
                  `sounds.ALL_TOKENS`.
    """
    return ''.join([L for L in phrase if L in valid])


def key_fn(s):
    """Sorting function for Sanskrit words in SLP1."""
    sa = "aAiIuUfFxXeEoOMHkKgGNcCjJYwWqQRtTdDnpPbBmyrlvSzsh '~"
    en = "123ABCDEFGHIJKLMNOPQRSTUVWabcdefghijklmnopqrstuvwxyz"
    mapper = dict(zip(sa, en))
    mapped = map(mapper.__getitem__, [x for x in s if x in ALL_SOUNDS])
    return ''.join(mapped)


# Letter transformations
# ----------------------

def letter_transform(name, docstring=None):
    data = {
        'shorten': dict(zip('AIUFX', 'aiufx')),
        'lengthen': dict(zip('aiufx', 'AIUFX')),
        'semivowel': dict(zip('iIuUfFxXeEoO',
                              'y y v v r r l l ay Ay av Av'.split())),
        'aspirate': dict(zip('kgcjwqtdpb',
                             'KGCJWQTDPB')),
        'deaspirate': dict(zip('KGCJWQTDPB',
                               'kgcjwqtdpb')),
        'voice': dict(zip('kKcCwWtTpP',
                          'gGjJqQdDbB')),
        'devoice': dict(zip('gGjJqQdDbB',
                            'kKcCwWtTpP')),
        'nasalize': dict(zip('kKgGhcCjJwWqQtTdDpPbB',
                             'NNNNNYYYYRRRRnnnnmmmm')),
        'dentalize': dict(zip('wWqQRz',
                              'tTdDns')),
        'retroflex': dict(zip('tTdDns',
                              'wWqQRz')),
        'simplify': dict(zip('kgGNhjtTdDpPbBnmsrH',
                             'kkkkkwttttppppnmHHH')),
        'guna': dict(zip('i I u U  f  F  x  X'.split(),
                         'e e o o ar ar al al'.split())),
        'vrddhi': dict(zip('a i I u U  f  F  x  X e o'.split(),
                           'A E E O O Ar Ar Al Al E O'.split())),
        'samprasarana': dict(zip('yrlv', 'ifxu'))
    }

    get = data[name].get

    def func(L):
        return get(L, L)

    if docstring is None:
        docstring = """{0} `L`. If this is not possible, return `L` unchanged.

        :param L: the letter to {1}
        """.format(name.capitalize(), name)

    func.__name__ = name
    func.__doc__ = docstring
    return func


shorten = letter_transform('shorten')
lengthen = letter_transform('lengthen')
semivowel = letter_transform('semivowel')
aspirate = letter_transform('aspirate')
deaspirate = letter_transform('deaspirate')
voice = letter_transform('voice')
devoice = letter_transform('devoice')
nasalize = letter_transform('nasalize')
dentalize = letter_transform('dentalize')
retroflex = letter_transform('retroflex')
simplify = letter_transform('simplify',
                            docstring="""
    Simplify the given letter, if possible.

    Here, to "simplify" a letter is to reduce it to a sound that is permitted
    to end a Sanskrit word. For instance, the `c` in `vAc` should be reduced
    to `k`::

        assert simplify('c') == 'k'

    :param letter: the letter to simplify
    """
                            )


guna = letter_transform('guna',
                        docstring="""
    Apply guna to the given letter, if possible.
    """
                        )


vrddhi = letter_transform('vrddhi',
                          docstring="""
    Apply vrddhi to the given letter, if possible.
    """
                          )


samprasarana = letter_transform('samprasarana',
                                docstring="""
    Apply samprasarana to the given letter, if possible.
    """
                                )

del letter_transform


# Term transformations
# --------------------

class Term(str):

    def simplify(self):
        """Simplify the given string using consonant reduction."""
        return self[:-1] + simplify(self[-1])


# Meter and metrical properties
# -----------------------------

def num_syllables(phrase):
    """Find the number of syllables in `phrase`.

    :param phrase: the phrase to test
    """
    return sum(1 for L in phrase if L in VOWELS)


def meter(phrase, heavy='_', light='.'):
    """Find the meter of the given phrase. Results are returned as a list
    whose elements are either `heavy` and `light`.

    By the traditional definition, a syllable is **heavy** if one of the
    following is true:

    - the vowel is long
    - the vowel is short and followed by multiple consonants
    - the vowel is followed by an anusvara or visarga

    All other syllables are **light**.

    :param phrase: the phrase to scan
    :param heavy: used to indicate heavy syllables. By default it's a string,
                  but you can pass in anything.
    :param light: used to indicate light syllables. By default it's a string,
                   but you can pass in anything.
    """
    scan = []
    had_consonant = False

    # True iff we've seen an anusvara, a visarga, or some conjunct consonants
    saw_cluster = False
    append = scan.append

    # Search for heavy syllable and call all other syllables light. Since
    # syllable weight can depend on later consonants, we have to look ahead
    # to determine the proper weight. An easy way to do that is to reverse
    # the string:
    for L in clean(phrase, ALL_SOUNDS)[::-1]:
        if L in VOWELS:
            if saw_cluster or L not in SHORT_VOWELS:
                append(heavy)
            else:
                append(light)

            saw_cluster = False

        elif L in 'MH' or had_consonant:
            saw_cluster = True
        had_consonant = L in CONSONANTS

    return scan[::-1]
